#include <string.h>
#include <stdlib.h>
#include "words.h"
#include "SHA3api_ref.h"
#include "hamsi.h"
#include "hamsi-tables.h"

HashReturn Init(hashState *state,int hashbitlen) {
    state->leftbits    = 0;
    state->counter     = 0;
    state->hashbitlen  = hashbitlen;
    state->ROUNDS      = 3*(hashbitlen/256);

    if (hashbitlen==224) { state->cvsize=256; memcpy(state->state,iv224,32); }
    else if (hashbitlen==256) { state->cvsize=256; memcpy(state->state,iv256,32); }
    else if (hashbitlen==384) { state->cvsize=512; memcpy(state->state,iv384,64); }
    else if (hashbitlen==512) { state->cvsize=512; memcpy(state->state,iv512,64); }
    else return BAD_HASHBITLEN;

    memset(state->leftdata,'\0',8);

    return SUCCESS;
}

HashReturn Update(hashState *state, const BitSequence *data, DataLength databitlen) {
    HashReturn ret=SUCCESS;
    int bits2hash=databitlen;
    int i;
    const BitSequence* block=data;
    const int s_blocksize=(state->cvsize/8);
    // Do first imcomplete block
    if (state->leftbits) {
        if ((state->leftbits+bits2hash) < s_blocksize) {
            // just copy, update leftbits
            for (i=0; i<((bits2hash+7)/8); ++i) state->leftdata[(state->leftbits/8)+i]=block[i];
            state->leftbits += bits2hash;
        } else {
            // copy and do first block from leftdata
            block += (s_blocksize - state->leftbits)/8;
            bits2hash -= (s_blocksize - state->leftbits);
        }
    }

    // do all complete blocks
    for (; bits2hash >= s_blocksize; bits2hash-=s_blocksize) {
        if (state->cvsize==256) {
            hash256(state->ROUNDS,state->state,block,0);
        } else if (state->cvsize==512) {
            hash512(state->ROUNDS,state->state,block,0);
        }
        else ret=BAD_HASHBITLEN;
        block += (s_blocksize/8);
        ++(state->counter);
    }


    // copy leftover data, update leftbits
    state->leftbits = bits2hash;
    for (i=0; i<((bits2hash+7)/8); ++i) state->leftdata[i]=block[i];

    return ret;
}

HashReturn Final(hashState* state, BitSequence *hashval) {
  const int block_bytes=(state->cvsize/8)/8;
  int i;
  U64 length = ((U64)(state->cvsize)*state->counter)+state->leftbits;
  BitSequence lenbytes[8];

  HashReturn ret=SUCCESS;

  // Padding
  state->leftdata[state->leftbits/8] &= (0xff<<(8-(state->leftbits%8)));
  state->leftdata[state->leftbits/8] |= (1<<(7-(state->leftbits%8)));
  for (i=(state->leftbits/8)+1; i<block_bytes; ++i) state->leftdata[i]=0;

  // Processing padding and length
  if (state->cvsize==256) {
      hash256(state->ROUNDS,state->state,state->leftdata,0);
      // position data length in lenbytes, little endian
      for (i=0; i<8; ++i) lenbytes[i] = 0xff&(length>>(8*i));
      hash256(state->ROUNDS,state->state,lenbytes,0);    if (ret) return ret;
      hash256(2*state->ROUNDS,state->state,lenbytes+4,1);  if (ret) return ret;
  } else if (state->cvsize==512) {
      hash512(state->ROUNDS,state->state,state->leftdata,0);
      // position data length in lenbytes, little endian
      for (i=0; i<8; ++i) lenbytes[i] = 0xff&(length>>(8*i));
      hash512(2*state->ROUNDS,state->state,lenbytes,1);   if (ret) return ret;
  }

  // T_384
  if (state->hashbitlen==384) {
      memcpy(hashval   ,state->state   , 8);
      memcpy(hashval+ 8,state->state+ 3,16);
      memcpy(hashval+24,state->state+ 8,12);
      memcpy(hashval+36,state->state+12, 8);
      memcpy(hashval+44,state->state+15, 4);
  } else memcpy(hashval,state->state,state->hashbitlen/8);

  return ret;
}

HashReturn Hash(int hashbitlen, const BitSequence* data, DataLength databitlen, BitSequence *hashval) {
    static hashState MyState;
    HashReturn ret;
    ret = Init(&MyState,hashbitlen);
    if (ret) return ret;
    ret = Update(&MyState,data,databitlen);
    if (ret) return ret;
    else return Final(&MyState,hashval);
}
