#include <string.h>
#include "hamsi.h"
#include "words.h"
#include "hamsi-tables.h"

inline void addConst(void *s,const void* c,int size) {
    int i=(size/64);
    do {
        --i;
        ((U64*)s)[i] ^= ((U64*)c)[i];
    } while (i);
}

#define X(n)    x[i]
#define S(n)    s[n*jump+i]

static Word x[8];

inline void subst(Word *s,const int jump) {
    int i;
    // From "Speeding up Serpent", by Dag Arne Osvik, University of Bergen"
    for (i=0; i<jump; ++i) {
        X(0) = S(0);	S(0) &= S(2);	
        S(0) ^= S(3);	S(2) ^= S(1);	
        S(2) ^= S(0);	S(3) |= X(0);	
        S(3) ^= S(1);	X(0) ^= S(2);	
        S(1) = S(3);	S(3) |= X(0);	
        S(3) ^= S(0);	S(0) &= S(1);	
        X(0) ^= S(0);	S(1) ^= S(3);	
        S(1) ^= X(0);	S(0) = ~X(0);	

        X(0) = S(2);
        S(2) = S(3);
        S(3) = X(0);
    }
}

#define ROTL32(v, n) (((v) << (n)) | ((v) >> (32 - (n))))

#define L(i,j,k,l)              \
    do {                        \
    s[i] = ROTL32(s[i],13);     \
    s[k] = ROTL32(s[k],3);      \
    s[j] ^= s[i] ^ s[k];        \
    s[l] ^= s[k] ^ (s[i]<<3);   \
    \
    s[j] = ROTL32(s[j],1);      \
    s[l] = ROTL32(s[l],7);      \
    s[i] ^= s[j] ^ s[l];        \
    s[k] ^= s[l] ^ (s[j]<<7);   \
    } while(0)

inline void diffuse256(Word s[16]) {
    L( 0, 5,10,15);
    L( 1, 6,11,12);
    L( 2, 7, 8,13);
    L( 3, 4, 9,14);
}

inline void diffuse512(Word s[32]) {
    L( 0, 9,18,27); L( 1,10,19,28);
    L( 2,11,20,29); L( 3,12,21,30);
    L( 4,13,22,31); L( 5,14,23,24);
    L( 6,15,16,25); L( 7, 8,17,26);

    L( 0, 2, 5, 7); L( 9,11,12,14);
    L(16,19,21,22); L(25,26,28,31);

    L( 3,10,17,24); L( 6,15,20,29);

    L( 1,13,18,30); L( 4, 8,23,27);
}

#define Exp(T,i) (T[0][d[0]][i]^T[1][d[1]][i]^T[2][d[2]][i]^T[3][d[3]][i])
#define dbg // do { cout << "[" << setfill('0') << hex; for (int g=0; g<16; ++g) cout << "." << setw(8) << s[g]; cout << endl; } while(0)
int hash256(const int ROUNDS, Word*cv, const Byte*d, int lastiter) {
    static Word s[16];
    int i;
    // Concatenation
    for (i=0; i<8; ++i) {
        s[4*((i+2)/4)+i] = Exp(T256,i);  // from message expansion
        s[4*(i/4)+i+2] = cv[i];          // from chain value
    }

    for (i=0; i<ROUNDS; ++i) {
        s[1] ^= i;
        addConst(s,a256[lastiter],512);
        subst(s,4);
        diffuse256(s);
    }

    // truncation
    cv[0] ^= s[0]; cv[1] ^= s[ 1]; cv[2] ^= s[ 2]; cv[3] ^= s[ 3];
    cv[4] ^= s[9]; cv[5] ^= s[10]; cv[6] ^= s[11]; cv[7] ^= s[13];

    return 0;
}

int hash512(const int ROUNDS, Word*cv, const Byte*d, int lastiter) {
    static Word s[32];
    int i;
    // Concatenation
    for (i=0; i<16; ++i) {
        s[2*((i/2)+((i/4)%2))+i] = Exp(T512,i);   // from message expansion
        s[2*((i/2)+(((i/4)+1)%2))+i] = cv[i];     // from chain value
    }

    for (i=0; i<ROUNDS; ++i) {
        s[1] ^= i;
        addConst(s,a512[lastiter],1024);
        subst(s,8);
        diffuse512(s);
    }

    // truncation
    for (i=0; i< 8; ++i) cv[i] ^= s[i];
    for (i=8; i<16; ++i) cv[i] ^= s[i+8];

    return 0;
}
